def expand_leading_tab(text: str, width: int = 4) -> str: ...
def expand_tab(text: str, space: str = "    ") -> str: ...
def escape(s: str, quote: bool = True) -> str: ...
def escape_url(link: str) -> str: ...
def safe_entity(s: str) -> str: ...
def unikey(s: str) -> str: ...
def unescape(s: str) -> str: ...
def striptags(s: str) -> str: ...
def strip_end(src: str) -> str: ...
