from ..markdown import Markdown

__all__ = ["table", "table_in_quote", "table_in_list"]

def table(md: Markdown) -> None: ...
def table_in_quote(md: Markdown) -> None: ...
def table_in_list(md: Markdown) -> None: ...
