from ..markdown import Markdown

__all__ = ["strikethrough", "mark", "insert", "superscript", "subscript"]

def strikethrough(md: Markdown) -> None: ...
def mark(md: Markdown) -> None: ...
def insert(md: Markdown) -> None: ...
def superscript(md: Markdown) -> None: ...
def subscript(md: Markdown) -> None: ...
